/*
 * AVC8000 Copyright (C) 2012 Advanced Micro Peripherals Ltd
 * Driver for AVC8000 cards
 *
 * (c) 2009-10 liran <jlee@techwellinc.com.cn> [Techwell China]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/videodev2.h>
#include <linux/delay.h>

#include <asm/div64.h>

#include "avc8000.h"
#include "avc8000-reg.h"
#include "avc8000-device.h"

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0))
#include "avc8000-videobuf-dma-contig.h"
#else
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,38))
#include <media/videobuf-dma-contig.h>
#else
#include "avc8000-videobuf-dma-contig.h"
#endif
#endif

int clkset(unsigned char (*Read)(struct avc8k_chip *chip,unsigned long  byte_addr), struct avc8k_chip *chip)
{
	int i,x=0;
	unsigned long y=0;
	for(i=0;i<4;i++)
		y+=Read(chip,0x06+i);
	for(i=0;i<3;i++)
		y+=Read(chip,0x10+i);
	for(i=0;i<16;i++)
		x=(x^((y>>i)&1));
		chip->vid_ch = x;
	return x;
}

static bool avc8k_check_286x(struct avc8k_chip *chip, u32 byte_addr)
{
    struct i2c_sw *i2c = &chip->i2c_686x.i2c;
	u8	data[3];

	data[0] = AVC8K2864_R_I2C_ADDR0;
	data[1] = (u8)(byte_addr&0xFF);								 //reg_addr

	if(!i2csw_write_bytes(i2c, data, 2)) {
		return false;
	}

	data[0] |= 1;												     //i2c rd-addr
	if(!i2csw_read_bytes(i2c, data, 1, &data[2], 1)) {
		return false;
	}

	return true;
}

//-------------------------------------------------------
// See avc8k_write_286x for description of byte_addr
//-------------------------------------------------------
static u8 avc8k_read_286x(struct avc8k_chip *chip, u32 byte_addr)
{
    struct i2c_sw *i2c = &chip->i2c_686x.i2c;
	u8	data[3];

	data[0] = AVC8K2864_R_I2C_ADDR0;
	data[1] = (u8)(byte_addr&0xFF);								 //reg_addr

	if(!i2csw_write_bytes(i2c, data, 2)) {
		return 0xFF;
	}

	data[0] |= 1;												     //i2c rd-addr
	if(!i2csw_read_bytes(i2c, data, 1, &data[2], 1)) {
		return 0xFF;
	}

	return data[2];
}

//-------------------------------------------------------
// See avc8k_write_286x for description of byte_addr
//-------------------------------------------------------
static bool avc8k_write_286x(struct avc8k_chip *chip, u32 byte_addr, u32 dat)
{
    struct i2c_sw *i2c = &chip->i2c_686x.i2c;
	u8	data[3];

	data[0] = AVC8K2864_R_I2C_ADDR0;
	data[1] = (u8)(byte_addr&0xFF);								 //reg_addr
	data[2] = (u8)(dat&0xFF);										 //reg_val

	return i2csw_write_bytes(i2c, data, 3);
}

int Tec_Init(struct avc8k_chip *chip)
{
	u32 Bit;
	Bit = DeviceReadRegister(chip,CLK_REG);
	Bit |= (1<<(16+7));
	Bit &= ~(1<<(16+6));
	Bit |= (1<<6);
	DeviceWriteRegister(chip,CLK_REG,Bit);

	Bit = DeviceReadRegister(chip,CLK_REG);

	if((Bit & (1<<7))==0)  //ch7
	{
		chip->vid_ch = 0;
		return 0;
	}
	Bit &= ~(1<<6);  //Set ch6
	DeviceWriteRegister(chip,CLK_REG,Bit);
	Bit = DeviceReadRegister(chip,CLK_REG);
	chip->vid_ch = (!(Bit&(1<<7)));
	return chip->vid_ch;
}

void GPIOI2C(struct avc8k_chip *chip, unsigned int GPIOaddr)
{
	//important ! Change GETACK also if changed control bits !
	unsigned int  GPIO_C_ADDR, control_word;

	control_word = 0;
	GPIO_C_ADDR = GPIOaddr;
	control_word = DeviceReadRegister(chip,GPIO_C_ADDR);
#ifdef FPGA
	control_word |= (1<<15);
#else
	control_word &= ~(1<<(AVC8K6864_I2C_SCLK_BIT+16));
	control_word &= ~(1<<(AVC8K6864_I2C_SDA_DATA_BIT+16));
	control_word |= (1<<AVC8K6864_I2C_SCLK_BIT);
	control_word |= (1<<AVC8K6864_I2C_SDA_DATA_BIT);
	DeviceWriteRegister(chip,GPIO_C_ADDR, control_word);
	control_word = DeviceReadRegister(chip,GPIO_C_ADDR);
#endif
}


int  avc8k_dev_init( struct avc8k_chip *chip )
{
	u32 tmp;
	int i = 0;

	u32 page_table_p[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE0, AVC8K6869_R_DMA_PAGE_TABLE2, AVC8K6869_R_DMA_PAGE_TABLE4, AVC8K6869_R_DMA_PAGE_TABLE6,
		AVC8K6869_R_DMA_PAGE_TABLE8, AVC8K6869_R_DMA_PAGE_TABLEA, AVC8K6869_R_DMA_PAGE_TABLEC, AVC8K6869_R_DMA_PAGE_TABLEE
	};
	u32 page_table_b[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE1, AVC8K6869_R_DMA_PAGE_TABLE3, AVC8K6869_R_DMA_PAGE_TABLE5, AVC8K6869_R_DMA_PAGE_TABLE7,
		AVC8K6869_R_DMA_PAGE_TABLE9, AVC8K6869_R_DMA_PAGE_TABLEB, AVC8K6869_R_DMA_PAGE_TABLED, AVC8K6869_R_DMA_PAGE_TABLEF
	};

	dprintk(DPRT_LEVEL0, chip, "%s()\n", __func__);

	if( pci_read_config_dword( chip->pci, 0x04, &tmp) == 0) {
		tmp |= 7;
		pci_write_config_dword( chip->pci, 0x04, tmp);
	}

	if( pci_read_config_dword( chip->pci, 0x78, &tmp) == 0) {
		tmp = (0x8<<5) | (tmp & 0xFFFFFE1F);		//8=128B, 9=256B, A=512B, B=1024B
		pci_write_config_dword( chip->pci, 0x78, tmp);
	}

	//Software Reset
	avc8k_write(AVC8K68XX_R_SYS_SOFT_RST, 0x01);
	avc8k_write(AVC8K68XX_R_SYS_SOFT_RST, 0x0F);

	for(i=0;i<AVC8K68XX_MAX_NUM_SG_DMA;i++){
	//S&G, Common Buffer DMA memory setup
//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE0, 0x00000000);   // -r 6869
//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE1, 0x00000000);	// -r 6869
	 avc8k_write(page_table_p[i], 0x00000000);
     avc8k_write(page_table_b[i], 0x00000000);


	}

	for(tmp=0; tmp<AVC8K68XX_MAX_NUM_DATA_DMA; tmp++) {
		avc8k_write(AVC8K68XX_R_CH8to15_CONFIG_REG_P(tmp),0);
		avc8k_write(AVC8K68XX_R_CH8to15_CONFIG_REG_B(tmp),0);
	}

	if( pci_read_config_dword(chip->pci, 0x04, &tmp) == 0) {
        dprintk(DPRT_LEVEL0,  chip, "enable INTx\n");
		tmp &= 0xFFFFFBFF;							// Enable INTx
		pci_write_config_dword(chip->pci, 0x04, tmp);
	}

	if( pci_read_config_dword(chip->pci, 0x50, &tmp) == 0) {	//CFG_MSI_CAP = 0x50
		tmp &= 0xFFFEFFFF;							//Disable MSI
		pci_write_config_dword( chip->pci, 0x50, tmp );
	}

	if( pci_read_config_dword(chip->pci, 0xAC, &tmp) == 0) {	//CFG_MSIX_CAP = 0xAC
		tmp &= 0x7FFFFFFF;							//Disable MSI-X
		pci_write_config_dword(chip->pci, 0xAC, tmp);
	}

	avc8k_write(AVC8K68XX_R_AVSRST,0x3F);             //Reset Internal video decoders
	avc8k_write(AVC8K68XX_R_DMA_CMD, 0);			     //Reset all DMA channels
	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, 0);	 //Disable all DMA channels
	avc8k_write(AVC8K68XX_R_DMA_INT_REF,  0x4c4b6);	 //Set INT interval for 2.0ms

	tmp = 0x00FF0004; //(AVC8K6864_MAXSWITCH>1) ? 0x00000004 : 0x00FF0004;
	avc8k_write(AVC8K68XX_R_DMA_CONFIG,  tmp);	    //enable vsync reset dma_operator fifo,INTx, MSI disabled,
												//little endian, bad/vsync protection on

	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_TIMEOUT,  0x14FF0FF0);//0x140c8b08);

	tmp  = 0x15DC;								//512*(2*125)/(27*720/858)
	avc8k_write(AVC8K68XX_R_PHASE_REF_CONFIG,  tmp);

	tmp = 1;
	tmp |= AVC8K68XX_R_VIDEO_CTRL1_STANDARDALL;	//pal, 4 D1 TDM format, 108MHz
	avc8k_write(AVC8K68XX_R_VIDEO_CTRL1,  tmp);


	//set gpio mode  //
	if(chip->pci->device == PCI_DEVICE_ID_AVC8K6869)  // for avc8000nano
	{
		tmp = avc8k_read(AVC8K6869_R_PIN_CFG_CTRL);
		tmp&= ~0x03;
		tmp|= 0x02;
		avc8k_write(AVC8K6869_R_PIN_CFG_CTRL,tmp);
	}

	tmp = 0x40a0;
	avc8k_write(AVC8K68XX_R_CSR_REG,tmp);

	tmp = AVC8K68XX_R_VIDEO_CTRL2_GENRSTALL | (AVC8K68XX_VIDEO_GEN_PATTERNS <<8) | AVC8K68XX_R_VIDEO_CTRL2_GENALL;
	avc8k_write(AVC8K68XX_R_VIDEO_CTRL2,  tmp);

	tmp = (AVC8K68XX_AUDIO_DMA_LENGTH << 19) |
		  (AVC8K68XX_AUDIO_SAMP_RATE_INT(AVC8K68XX_AUDIO_SAMPLE_RATE)<<5) |
		  (AVC8K68XX_AUDIO_GEN_PATTERN << 4) |
		  (AVC8K68XX_AUDIO_GEN_MIX_SEL <<1 ) |
		   AVC8K68XX_AUDIO_GEN;
	avc8k_write(AVC8K68XX_R_AUDIO_CTRL1,  tmp);

	tmp = (u32)AVC8K68XX_AUDIO_SAMP_RATE_EXT(AVC8K68XX_AUDIO_SAMPLE_RATE);
	avc8k_write(AVC8K68XX_R_AUDIO_CTRL2,  tmp);

	tmp = 0x40;
	if(AVC8K68XX_AUDIO_SAMPLE_BIT == 8) {
		tmp |= BIT(8);
	}
	avc8k_write(AVC8K68XX_R_AUDIO_CTRL3, tmp);


	//Show Blue background if no signal
//	if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)  // for avc8000x
//	{
		avc8k_write(AVC8K68XX_R_MISC_CONTROL2, 0xE0);//0xE7);
//	else
//		avc8k_write(TW6864_R_MISC_CTRLII1, 0xE0);//0xE7);
//	}


	tmp = 0x20;

	if(chip->pci->device == PCI_DEVICE_ID_AVC8K6869)  // for avc8000nano
	{
		tmp = 0x40;
		avc8k_write(AVC8K6869_R_AUDIO_GAIN_4, tmp);
		avc8k_write(AVC8K6869_R_AUDIO_GAIN_5, tmp);
		avc8k_write(AVC8K6869_R_AUDIO_GAIN_6, tmp);
		avc8k_write(AVC8K6869_R_AUDIO_GAIN_7, tmp);
	}

	avc8k_write(AVC8K68XX_R_AUDIO_GAIN_0, tmp);
	avc8k_write(AVC8K68XX_R_AUDIO_GAIN_1, tmp);
	avc8k_write(AVC8K68XX_R_AUDIO_GAIN_2, tmp);
	avc8k_write(AVC8K68XX_R_AUDIO_GAIN_3, tmp);

	avc8k_write(AVC8K68XX_R_VERTICAL_CTRL,0x00); //0x26 will cause ch0 and ch1 have dma_error.
	avc8k_write(AVC8K68XX_R_LOOP_CTRL,    0xa5);
	avc8k_write(AVC8K68XX_R_GPIO_REG,	    0xFFFF0000);
 	avc8k_write(AVC8K68XX_R_HFLT1,		0x00);//0xBB);		//NOTE: //NOTE: This will affect the clarity of CIF to be set for CIF & 4CIF
 	avc8k_write(AVC8K68XX_R_HFLT2,		0x00);//0xBB);

//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE0, (u32)chip->dma_desc[AVC8K68XX_DMA_DESC_P].dma);
//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE1, (u32)chip->dma_desc[AVC8K68XX_DMA_DESC_B].dma);

    return 0;
}

void avc8k_dev_uninit( struct avc8k_chip *chip )
{
	int   i = 0;
	u32 page_table_p[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE0, AVC8K6869_R_DMA_PAGE_TABLE2, AVC8K6869_R_DMA_PAGE_TABLE4, AVC8K6869_R_DMA_PAGE_TABLE6,
		AVC8K6869_R_DMA_PAGE_TABLE8, AVC8K6869_R_DMA_PAGE_TABLEA, AVC8K6869_R_DMA_PAGE_TABLEC, AVC8K6869_R_DMA_PAGE_TABLEE
	};
	u32 page_table_b[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE1, AVC8K6869_R_DMA_PAGE_TABLE3, AVC8K6869_R_DMA_PAGE_TABLE5, AVC8K6869_R_DMA_PAGE_TABLE7,
		AVC8K6869_R_DMA_PAGE_TABLE9, AVC8K6869_R_DMA_PAGE_TABLEB, AVC8K6869_R_DMA_PAGE_TABLED, AVC8K6869_R_DMA_PAGE_TABLEF
	};
	dprintk(DPRT_LEVEL0, chip, "%s()\n", __func__);

    avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, 0);	 //Disable all DMA channels //070514
	avc8k_write(AVC8K68XX_R_DMA_CMD, 0);			     //Reset all DMA channels
	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, 0);	 //Disable all DMA channels


//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE0, 0x00000000);
//	avc8k_write(AVC8K68XX_R_DMA_PAGE_TABLE1, 0x00000000);
	for(i=0; i<AVC8K68XX_MAX_NUM_SG_DMA; i++) {
		avc8k_write(AVC8K68XX_R_VC_CTRL_REG(i), 0x00000000);
		avc8k_write(page_table_p[i], 0x00000000);
		avc8k_write(page_table_b[i], 0x00000000);
	}

}

//-------------------------------------------------------
//Initialize 286x
//-------------------------------------------------------
int avc8k_dev_init_286x( struct avc8k_chip *chip )
{
	u32	tmp;

	tmp  = avc8k_read(AVC8K68XX_R_CSR_REG);
	tmp &= 0x7FFF;
	avc8k_write(AVC8K68XX_R_CSR_REG,tmp);
	msleep(100);
	tmp |= 0x8002;  // Pull out from reset and enable I2S
	avc8k_write(AVC8K68XX_R_CSR_REG,tmp);
	GPIOI2C(chip, CLK_REG);

	// 2864 #1/External, Muxed-656
	tmp  = 0x02;
	avc8k_write_286x(chip, AVC8K2864_R_CLK_SEL_2865, tmp);	// if outside is TW2865 instead of TW2864
	tmp = avc8k_read_286x(chip, AVC8K2864_R_CLK_SEL_2865);
	dprintk(1, chip, "%s() 286[CLK_SEL_2865 0x%x]=0x%x\n", __func__, AVC8K2864_R_CLK_SEL_2865, tmp);

	if(!avc8k_check_286x(chip, AVC8K2864_R_CLK_SEL_2865)) {
	    return 0;
	}

	tmp = 0x73;					// CHID with 656 Sync code, 656 output even no video input
	if(!avc8k_write_286x(chip, AVC8K2864_R_NOVID,tmp)) {
		return 0;
	}

	tmp = 0x80;					// Serial mode
	avc8k_write_286x(chip, AVC8K2864_R_SMD,tmp);

	tmp = 0xC0;					// I2S,8-bit,ASYNR is output,Slave
	if(AVC8K68XX_AUDIO_SAMPLE_BIT == 8) {
		tmp |= BIT(2);
	}
	else {
		tmp &= ~BIT(2);
	}
	avc8k_write_286x(chip, AVC8K2864_R_MASTER_CTRL,tmp);

	tmp = 0x33;					// Audio Gain
	avc8k_write_286x(chip, AVC8K2864_R_AUDIO_GAIN0,tmp);

	tmp = 0x33;					// Audio Gain
	avc8k_write_286x(chip, AVC8K2864_R_AUDIO_GAIN1,tmp);

	tmp  = 0x1;
	avc8k_write_286x(chip, AVC8K2864_R_NUM_ADUIOR,tmp);

	tmp  = 0x32;
	avc8k_write_286x(chip, AVC8K2864_R_SEQ_ADUIOR,tmp);

	tmp =  2 | (2<<2) | (2<<4) | (2<<6);
	avc8k_write_286x(chip, AVC8K2864_R_VIDEO_OUT_CTRL,tmp);

	tmp = 0x40 | ( 2 | (2<<2) ) ; // Out enable
	avc8k_write_286x(chip, AVC8K2864_R_CLKOCTRL,tmp);
	tmp=avc8k_read_286x(chip, AVC8K2864_R_CLKOCTRL);
	dprintk(1, chip, "%s() 2864[CLKOCTRL 0x%x]=0x%x\n", __func__,AVC8K2864_R_CLKOCTRL,tmp);

	avc8k_write_286x(chip, 0x6A,0x0F);
	avc8k_write_286x(chip, 0x6B,0x0F);
	avc8k_write_286x(chip, 0x6C,0x0F);
	avc8k_write_286x(chip, 0x60,0x45);
	avc8k_write_286x(chip, 0x4B,0x30);

	avc8k_write_286x(chip, AVC8K2864_R_LOOP_CTRL, 0xa5);
	avc8k_write_286x(chip, AVC8K2864_R_VERTICAL_CTRL, 0x00);
	avc8k_write_286x(chip, AVC8K2864_R_VERTICAL_SHARPNESS, 0x80);
 	avc8k_write_286x(chip, AVC8K2864_R_HFLT1, 0x00);//0xBB);		//NOTE: This will affect the clarity of CIF to be set for CIF & 4CIF
 	avc8k_write_286x(chip, AVC8K2864_R_HFLT2, 0x00);//0xBB);

 	return AVC8K68XX_DECODER_COUNT(chip->pci->device);
}

void avc8k_dev_setdma( struct avc8k_chip *chip, u32 dma_cmd)
{
	u32  dma_en = dma_cmd;

	if(dma_cmd != 0) {
		dma_en |= (1<<31);
	}

	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, dma_en);
	avc8k_write(AVC8K68XX_R_DMA_CMD, dma_cmd);
}

int avc8k_dev_set_mux(struct avc8k_vdev *dev, unsigned int input)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
	u32 uConfig = avc8k_read(AVC8K68XX_R_CH0to7_CONFIG_REG(ch));

	dvprintk(DPRT_LEVEL0, dev, "%s(%d)\n", __func__, (int)input);

	uConfig &= ~(3 << 30);
	uConfig |= (input&3) << 30;
	avc8k_write(AVC8K68XX_R_CH0to7_CONFIG_REG(ch), uConfig);

	if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)  // for avc8000x
	{
		if(ch >= AVC8K68XX_DECODER_COUNT(dev->chip->pci->device)) {
			ch -= AVC8K68XX_DECODER_COUNT(dev->chip->pci->device);
			uConfig = avc8k_read_286x(chip, AVC8K2864_R_VERTICAL_SHARPNESS);
			uConfig&= ~(1<<ch);
			uConfig|= (input%AVC8K2865_MUX_COUNT) << ch;
			avc8k_write_286x(chip, AVC8K2864_R_VERTICAL_SHARPNESS, (u8)uConfig);
		}
		if(Tec_Init(chip)==0)
			clkset(&DeviceWriteFPGA1,chip);
	}
	else
	{
		clkset(&DeviceReadFPGA1,chip);
	}

	return 0;
}

void avc8k_dev_set_standard(struct avc8k_vdev *dev, v4l2_std_id value, bool b_auto)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
	u32 tmp = avc8k_read(AVC8K68XX_R_VIDEO_CTRL1);

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	tmp &= ~AVC8K68XX_R_VIDEO_CTRL1_STANDARD(ch);
	if(IS_PAL(value)) {
		tmp |= AVC8K68XX_R_VIDEO_CTRL1_STANDARD(ch);		//pal, 4 D1 TDM format, 108MHz
	}

	avc8k_write(AVC8K68XX_R_VIDEO_CTRL1,  tmp);

	if(ch < 4) {
		if(b_auto) {
			tmp = 0x07;
		}
		else {
			tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_STANDARD_0, ch));
			tmp&= 0xF8;
			tmp|= IS_PAL(value) ? 1 : 0;
		}
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_STANDARD_0, ch), tmp);
	}
	else
	{
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			ch -= 4;
			if(b_auto) {
				tmp = 0x07;
			}
			else
			{
				tmp = avc8k_read_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_STANDARD_0, ch));
				tmp&= 0xF8;
				tmp|= IS_PAL(value) ? 1 : 0;
			}
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_STANDARD_0, ch), tmp);
		}
		else   // 6869
		{
			ch -= 4;
				if(b_auto) {
					tmp = 0x07;
				}
				else
				{
					tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K6869_R_STANDARD_4, ch));
					tmp&= 0xF8;
					tmp|= IS_PAL(value) ? 1 : 0;
				}
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_STANDARD_4, ch), tmp);
		}
	}
}

static u32 avc8k_dev_map_pixelformat(u32 fourcc)
{
	switch(fourcc) {
	case V4L2_PIX_FMT_YUYV :
		fourcc = (AVC8K68XX_VIDEO_FORMAT_YUYV & 7);
		break;
	case V4L2_PIX_FMT_UYVY :
		fourcc = (AVC8K68XX_VIDEO_FORMAT_UYVY & 7);
		break;
	case V4L2_PIX_FMT_YVU420 :
		fourcc = (AVC8K68XX_VIDEO_FORMAT_YUV420 & 7);
		break;
	case V4L2_PIX_FMT_YUV411P :
		fourcc = (AVC8K68XX_VIDEO_FORMAT_Y411 & 7);
		break;
	case V4L2_PIX_FMT_Y41P :
		fourcc = (AVC8K68XX_VIDEO_FORMAT_Y41P & 7);
		break;
	case V4L2_PIX_FMT_RGB565:
		fourcc = (AVC8K68XX_VIDEO_FORMAT_RGB565 & 7);
		break;
	case V4L2_PIX_FMT_RGB555:
		fourcc = (AVC8K68XX_VIDEO_FORMAT_RGB555 & 7);
		break;
	default:
		fourcc = (AVC8K68XX_VIDEO_FORMAT_YUYV & 7);
	}

	return fourcc;
}

void avc8k_dev_set_pixelformat(struct avc8k_vdev *dev, u32 fourcc)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
	u32 uConfig = avc8k_read(AVC8K68XX_R_CH0to7_CONFIG_REG(ch));

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	uConfig &= ~(7 << 20);
	uConfig |= avc8k_dev_map_pixelformat(fourcc) << 20;
	avc8k_write(AVC8K68XX_R_CH0to7_CONFIG_REG(ch), uConfig);
}

void avc8k_dev_set_size(struct avc8k_vdev *dev, int width, int height)
{
    struct avc8k_chip *chip = dev->chip;
    int   ch = dev->channel_id;
	u32   tmp, tmp1, tmp2;
	int   maxWidth = AVC8K68XX_REMOVE_BLACKSTRIPE ? 720 : 704;
	bool  isPal = IS_PAL(dev->tvnorm);

	dvprintk(DPRT_LEVEL0, dev, "%s(%d, %d)\n", __func__, width, height);

	//SH_Scaler
	//setup for Black stripe remover
	tmp1 = width-12;   //EndPos
	tmp2 = 4;		   //StartPos
	tmp = (tmp1 - tmp2)*(1<<16)/width;
	tmp = (tmp2 & 0x1F) | ((tmp1 & 0x3FF) << 5) | (tmp <<15);
	tmp |= AVC8K68XX_REMOVE_BLACKSTRIPE;
	avc8k_write(AVC8K68XX_R_SHSCALER_REG(ch), tmp);

	tmp = (u32)AVC8K68XX_VIDEO_SIZE(width,height);
	tmp1 = tmp2 = tmp;
	avc8k_write(AVC8K68XX_R_VIDEO_SIZE_REGA, tmp);
	avc8k_write(AVC8K68XX_R_VIDEO_SIZE_REG(ch), tmp); //for Rev.B or later only

	//H/V Scale
	tmp1 &= 0x7FF;
	tmp2 = (tmp2>>16)&0x1FF;	 //V
	tmp1 = (maxWidth*256)/tmp1;  //H
	tmp2 = ((isPal?288:240)*256)/tmp2;
	tmp = tmp1 & 0xFF;

	if(ch < 4) {
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HSCALE_LO_0,ch),  tmp);

		tmp = tmp2 & 0xFF;
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VSCALE_LO_0,ch),  tmp);

		tmp = (((tmp2 >> 8)& 0xF) << 4) | ( (tmp1>>8) & 0xF );
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VHSCALE_HI_0,ch),  tmp);

		if(tmp1 > 1) {
			if(width==352)		{ tmp = 0x01; }
			else if(width==176)	{ tmp = 0x02; }
			else				{ tmp = 0x00; }
		}
		else					{ tmp = 0x0B; }
		tmp1 = avc8k_read(AVC8K68XX_R_HFLT1 + (ch/2)*4);
		tmp1&= ~(0x0F << ((ch%2)*4));
		tmp1|= ((tmp&0x0F) << ((ch%2)*4));
		avc8k_write(AVC8K68XX_R_HFLT1 + (ch/2)*4, tmp1);

		if(isPal) {
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_0,ch), 0x17);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HDELAY_0,ch), AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x00E);
		}
		else {
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_0,ch), 0x15/*0x14*/);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HDELAY_0,ch), 0x0E);//AVC8K6864_REMOVE_BLACKSTRIPE?0x0A:0x00C);
		}
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_F2_CNT_0,ch), 0x00);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_HSCALE_LO_0+ch*3,tmp);

			tmp = tmp2 & 0xFF;
			avc8k_write_286x(chip, AVC8K2864_R_VSCALE_LO_0+ch*3,tmp);

			tmp = (((tmp2 >> 8)& 0xF) << 4) | ( (tmp1>>8) & 0xF );
			avc8k_write_286x(chip, AVC8K2864_R_VHSCALE_HI_0+ch*3,tmp);

			if(tmp1 > 1) {
				if(width==352)		{ tmp = 0x01; }
				else if(width==176)	{ tmp = 0x02; }
				else				{ tmp = 0x00; }
			}
			else					{ tmp = 0x0B; }
			tmp1 = avc8k_read_286x(chip, AVC8K2864_R_HFLT1 + ch/2);
			tmp1&= ~(0x0F << ((ch%2)*4));
			tmp1|= ((tmp&0x0F) << ((ch%2)*4));
			avc8k_write_286x(chip, AVC8K2864_R_HFLT1 + ch/2, tmp1);

			if(isPal) {
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_VDELAY_0,ch), 0x17);
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_HDELAY_0,ch), AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x0E);
			}
			else {
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_VDELAY_0,ch), 0x15/*0x14*/);
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_HDELAY_0,ch), 0x0E);//AVC8K6864_REMOVE_BLACKSTRIPE?0x0A:0x00C);
			}


		}
		else  // for 6869
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HSCALE_LO_4,ch),tmp);

			tmp = tmp2 & 0xFF;
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VSCALE_LO_4,ch),tmp);

			tmp = (((tmp2 >> 8)& 0xF) << 4) | ( (tmp1>>8) & 0xF );
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VHSCALE_HI_4,ch),tmp);

			if(tmp1 > 1) {
				if(width==352)		{ tmp = 0x01; }
				else if(width==176)	{ tmp = 0x02; }
			else				{ tmp = 0x00; }
			}
			else					{ tmp = 0x0B; }
			tmp1 = avc8k_read(AVC8K6869_R_HFLT1 + (ch/2)*4);
			tmp1&= ~(0x0F << ((ch%2)*4));
			tmp1|= ((tmp&0x0F) << ((ch%2)*4));
			avc8k_write(AVC8K6869_R_HFLT1 + (ch/2)*4, tmp1);


			if(isPal) {
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_4,ch), 0x17);
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HDELAY_4,ch), AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x0E);
			}
			else {
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_4,ch), 0x15/*0x14*/);
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HDELAY_4,ch), 0x0E);//AVC8K6864_REMOVE_BLACKSTRIPE?0x0A:0x00C);
			}
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_F2_CNT_4,ch), 0x00);
		}
	}
}

void avc8k_dev_set_size2(struct avc8k_vdev *dev, int width, int height)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
	u32 tmp, tmp1, tmp2;
	int maxWidth = AVC8K68XX_REMOVE_BLACKSTRIPE ? 720 : 704;

	dvprintk(DPRT_LEVEL0, dev, "%s(%d, %d)\n", __func__, width, height);

	tmp = (u32)AVC8K68XX_VIDEO_SIZE_F2(width,height);
	tmp1 = tmp2 = tmp;
	avc8k_write(AVC8K68XX_R_VIDEO_SIZE_REG_F2(ch), tmp); //for Rev.B or later only

	if(ch < 4) {
		//H/V Scale
		tmp1&= 0x7FF;
		tmp2 = (tmp2>>16)&0x1FF;	 //V
		tmp1 = (maxWidth*256)/tmp1;  //H
		tmp2 = ((IS_PAL(dev->tvnorm)?288:240)*256)/tmp2;
		tmp  = tmp1 & 0xFF;

		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HSCALE_LO_F2_0,ch),  tmp);

		tmp = tmp2 & 0xFF;
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VSCALE_LO_F2_0,ch),  tmp);

		tmp = (((tmp2 >> 8)& 0xF) << 4) | ( (tmp1>>8) & 0xF );
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VHSCALE_HI_F2_0,ch),  tmp);

		if(IS_PAL(dev->tvnorm)) {
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_F2_0,ch), 0x17);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HDELAY_F2_0,ch), AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x00E);
		}
		else {
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_F2_0,ch), 0x14);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HDELAY_F2_0,ch), 0x0E);//AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x00C);
		}

		tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_CROPPING_0,ch));
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_CROPPING_F2_0,ch), tmp);

		tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HACTIVE_0,ch));
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_HACTIVE_F2_0,ch), tmp);

		tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VACTIVE_0,ch));
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VACTIVE_F2_0,ch), tmp);

		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_F2_CNT_0,ch), 0x1);
	}
	else
		if (chip->pci->device == PCI_DEVICE_ID_AVC8K6869)  // for avc8000nano
		{
				ch -= 4;
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HSCALE_LO_F2_4,ch),  tmp);

				tmp = tmp2 & 0xFF;
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VSCALE_LO_F2_4,ch),  tmp);

				tmp = (((tmp2 >> 8)& 0xF) << 4) | ( (tmp1>>8) & 0xF );
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VHSCALE_HI_F2_4,ch),  tmp);

				if(IS_PAL(dev->tvnorm)) {
					avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_F2_4,ch), 0x17);
					avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HDELAY_F2_4,ch), AVC8K68XX_REMOVE_BLACKSTRIPE?0x0A:0x00E);
				}
				else {
					avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_F2_4,ch), 0x14);
					avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HDELAY_F2_4,ch), 0x0E);//TW686X_REMOVE_BLACKSTRIPE?0x0A:0x00C);
				}

				tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K6869_R_CROPPING_4,ch));
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_CROPPING_F2_4,ch), tmp);

				tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HACTIVE_4,ch));
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HACTIVE_F2_4,ch), tmp);

				tmp = avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VACTIVE_4,ch));
				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VACTIVE_F2_4,ch), tmp);

				avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_F2_CNT_4,ch), 0x1);

		}
		tmp  = avc8k_read(AVC8K68XX_R_VIDEO_SIZE_REG(ch));
		tmp |= (1 << 30);
		avc8k_write(AVC8K68XX_R_VIDEO_SIZE_REG(ch), tmp);
}

void avc8k_dev_set_vdelay(struct avc8k_vdev *dev, int value)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch < 4) {
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_0,ch), (u32)value);
		avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VDELAY_F2_0,ch), (u32)value);
	}
	else
	{
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_VDELAY_0,ch), (u32)value);
		}
		else  // 6869
		{
			ch -= 4;
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_4,ch), (u32)value);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VDELAY_F2_4,ch), (u32)value);
		}
	}
}

void avc8k_dev_set_bright(struct avc8k_vdev *dev, int val)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	//if(1) { val = 0x00; }

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch<4) {
		avc8k_write(AVC8K68XX_R_BRIGHT(ch), val);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_BRIGHT_0,ch), val);
		}
		else
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_BRIGHT_4,ch), val);
		}
	}
}

void avc8k_dev_set_contrast(struct avc8k_vdev *dev, int val)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	//if(1) { val = 0x64; }

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch<4) {
		avc8k_write(AVC8K68XX_R_CONTRAST(ch),  val);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_CONTRAST_0,ch), val);
		}
		else //6869
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_CONTRAST_4,ch), val);
		}
	}
}

void avc8k_dev_set_hue(struct avc8k_vdev *dev, int val)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	//if(1) { val = 0; }

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch<4) {
		avc8k_write(AVC8K68XX_R_HUE(ch),  val);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_HUE_0,ch), val);
		}
		else // 6869
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_HUE_4,ch), val);
		}
	}
}

void avc8k_dev_set_saturation(struct avc8k_vdev *dev, int val)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	//if(1) { val = 0x80; }

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch<4) {
		avc8k_write(AVC8K68XX_R_SAT_U(ch),  val);
		avc8k_write(AVC8K68XX_R_SAT_V(ch),  val);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_SAT_U_0,ch), val);
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_SAT_V_0,ch), val);
		}
		else // 6869
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_SAT_U_4,ch), val);
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_SAT_V_4,ch), val);
		}
	}
}

void avc8k_dev_set_sharpness(struct avc8k_vdev *dev, int val)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	val &= 0x0F;
	val |= 0x10;//0xA0;

	//if(1) { val = 0x11; }

	dvprintk(DPRT_LEVEL0, dev, "%s()\n", __func__);

	if(ch<4) {
		avc8k_write(AVC8K68XX_R_SHARPNESS(ch),  val);
	}
	else {
		ch -= 4;
		if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
		{
			avc8k_write_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_SHARPNESS_0,ch), val);
		}
		else // 6869
		{
			avc8k_write(AVC8K68XX_R_CHANNEL(AVC8K6869_R_SHARPNESS_4,ch), val);
		}
	}
}

u32 avc8k_dev_get_decoderstatus(struct avc8k_vdev *dev)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;

	if(ch < 4) {

		return avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K68XX_R_VSTATUS_0, ch));
	}

	if(chip->pci->device != PCI_DEVICE_ID_AVC8K6869)
	{
		return avc8k_read_286x(chip, AVC8K2864_R_CHANNEL(AVC8K2864_R_VSTATUS_0, ch-4));
	}
	else //6869
	{
		return avc8k_read(AVC8K68XX_R_CHANNEL(AVC8K6869_R_VSTATUS_4, ch-4));
	}
}

int avc8k_dev_get_signal(struct avc8k_vdev *dev)
{
    u32 vstatus = dev->vstatus; //avc8k_dev_get_decoderstatus( dev );

	dvprintk(DPRT_LEVEL0, dev, "%s() vstatus=%x\n", __func__, vstatus);

    return (!(vstatus&BIT(7)) && (vstatus&BIT(6)));
}

u32 avc8k_dev_map_fieldrate(int nFieldRate, bool bNTSC)
{
	u32 uRate = 0;

	if(!bNTSC) {
		switch(nFieldRate) {
		case 0:
			uRate = 0;
			break;
		case 1:
		case 2:
			uRate = 0x00000001;
			break;
		case 3:
			uRate = 0x00004001;//0x10001000;
			break;
		case 4:
		case 5:
			uRate = 0x00104001;//0x10010010;
			break;
		case 6:
		case 7:
			uRate = 0x00404041;//0x10101010;
			break;
		case 8:
		case 9:
			uRate = 0x01041041;//0x11011010;
			break;
		case 10:
			uRate = 0x01104411;//0x11101110;
			break;
		case 11:
		case 12:
			uRate = 0x01111111;//0x11111110;
			break;
		case 13:
		case 14:
			uRate = 0x04444445;//0x11111111;
			break;
		case 15:
			uRate = 0x04511445;//0x15111111;
			break;
		case 16:
		case 17:
			uRate = 0x05145145;//0x15111511;
			break;
		case 18:
		case 19:
			uRate = 0x05151515;//0x15151511;
			break;
		case 20:
			uRate = 0x05515455;//0x15151515;
			break;
		case 21:
		case 22:
			uRate = 0x05551555;//0x15515515;
			break;
		case 23:
		case 24:
			uRate = 0x05555555;//0x15551555;
			break;
		default:
			uRate = 0x15555555;
			break;
		}
	}
	else {
		switch(nFieldRate) {
		case 0:
			uRate = 0;
			break;
		case 1:
		case 2:
		case 3:
			uRate = 0x00000001;
			break;
		case 4:
		case 5:
			uRate = 0x00004001;//0x10001000;
			break;
		case 6:
		case 7:
			uRate = 0x00104001;//0x10010010;
			break;
		case 8:
		case 9:
			uRate = 0x00404041;//0x10101010;
			break;
		case 10:
		case 11:
			uRate = 0x01041041;//0x11011010;
			break;
		case 12:
		case 13:
			uRate = 0x01104411;//0x11101110;
			break;
		case 14:
		case 15:
			uRate = 0x01111111;//0x11111110;
			break;
		case 16:
		case 17:
			uRate = 0x04444445;//0x11111111;
			break;
		case 18:
		case 19:
			uRate = 0x04511445;//0x15111111;
			break;
		case 20:
		case 21:
			uRate = 0x05145145;//0x15111511;
			break;
		case 22:
		case 23:
			uRate = 0x05151515;//0x15151511;
			break;
		case 24:
		case 25:
			uRate = 0x05515455;//0x15151515;
			break;
		case 26:
		case 27:
			uRate = 0x05551555;//0x15551555;
			break;
		case 28:
		case 29:
			uRate = 0x05555555;//0x15551555;
			break;
		default:
			uRate = 0x15555555;
			break;
		}
	}

	return uRate;
}

u32 avc8k_dev_map_framerate(int nFieldRate1, int nFieldRate2, bool bNTSC, bool b6864)
{
	u32 uRate = 0;

	if(nFieldRate1 == 0) {
		uRate = avc8k_dev_map_fieldrate(nFieldRate2, bNTSC);
	}
	else if(nFieldRate2 == 0) {
		uRate = (avc8k_dev_map_fieldrate(nFieldRate1, bNTSC) << 1);
	}
	else {
		if(b6864) {
			uRate = bNTSC ? 30 : 25;
			if(nFieldRate1>(int)uRate) nFieldRate1 = (int)uRate;
			if(nFieldRate2>(int)uRate) nFieldRate2 = (int)uRate;
			if(uRate==nFieldRate1) {
				uRate = avc8k_dev_map_fieldrate(nFieldRate1, bNTSC) |
						(avc8k_dev_map_fieldrate(nFieldRate2, bNTSC) << 1);
			}
			else {
				uRate = (avc8k_dev_map_fieldrate(nFieldRate1, bNTSC)<<1) |
						(avc8k_dev_map_fieldrate(nFieldRate2, bNTSC)<<2);
			}
		}
		else {
			uRate = avc8k_dev_map_fieldrate(nFieldRate1, bNTSC) |
					(avc8k_dev_map_fieldrate(nFieldRate2, bNTSC) << 1);
		}
	}
	if(uRate > 0) {
		uRate |= 0x80000000;
	}

	return uRate;
}

void avc8k_dev_set_framerate(struct avc8k_vdev *dev, int val, int field)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
    u32 uRate = 0;

    if(field == V4L2_FIELD_TOP) {
        uRate = avc8k_dev_map_framerate(val, 0, !IS_PAL(dev->tvnorm), false);
    }
    else if(field == V4L2_FIELD_BOTTOM) {
        uRate = avc8k_dev_map_framerate(0,val, !IS_PAL(dev->tvnorm), false);
    }
    else {
        uRate = avc8k_dev_map_framerate(val, val, !IS_PAL(dev->tvnorm), (dev->channel_id<4));
    }

	avc8k_write(AVC8K68XX_R_DROP_FIELD_REG(ch), uRate); //0x95555555 for even only, 0xd5555555 for odd only
}

//----------------------------------------------
// Start/Stop DMA Program
//----------------------------------------------
void avc8k_dev_run(struct avc8k_vdev *dev, bool b_run)
{
    struct avc8k_chip *chip = dev->chip;
    int  ch   = dev->channel_id;
	u32  tmp  = avc8k_read(AVC8K68XX_R_DMA_CMD);
	u32  tmp1 = avc8k_read(AVC8K68XX_R_DMA_CHANNEL_ENABLE);
	u32 tmpstatus = avc8k_dev_get_decoderstatus(dev);
	int matchedvideo=1;

	dvprintk(DPRT_LEVEL0, dev, "%s(%d %d)\n", __func__, ch, b_run);

	if((tmpstatus&0x00000048)==0x00000048)
	{
		if(IS_PAL(dev->tvnorm) && !(tmpstatus&0x01))
		{
			dvprintk(DPRT_LEVEL0, dev, "%s: Expected PAL but not 50Hz\n", __func__);
			matchedvideo = 0;
		}
		if(!IS_PAL(dev->tvnorm) && (tmpstatus&0x01))
		{
			dvprintk(DPRT_LEVEL0, dev, "%s: Expected NTSC but not 60Hz\n", __func__);
			matchedvideo = 0;
		}
	}
	else
	{
		matchedvideo = 0;
	}

	dvprintk(DPRT_LEVEL0, dev, "%s: tmpstatus=%x %d\n", __func__, tmpstatus, matchedvideo);

	if(matchedvideo)
	{
	if(b_run) {
        tmp |= (chip->vid_ch<<31);
        tmp |= (chip->vid_ch<<ch);
		tmp1|= (chip->vid_ch<<ch);
	}
	else {
		if(!(tmp&(1<<ch)) && !(tmp1&(1<<ch))) {
			return;
		}
		tmp &= ~(1<<ch);
		tmp1&= ~(1<<ch);
		if( tmp1 == 0 ) {
		    tmp = 0;
		}
	}

	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, tmp1);
	avc8k_write(AVC8K68XX_R_DMA_CMD, tmp);
	}
}

void avc8k_dev_set_dma(struct avc8k_vdev *dev, int width, int height, int linewidth)
{
	struct avc8k_dmadesc * dma_desc;
    struct avc8k_chip *chip = dev->chip;
    int   ch = dev->channel_id;
	bool  is4CIF = IS_PAL(dev->tvnorm) ? (height>288) : (height>240);
	int   nHeight = !is4CIF ? height : height/2;
	u32   uTemp, uConfig;
	int   linepitch = linewidth;
	int   idx_begin = ch*AVC8K68XX_DMA_DESC_UNIT;
	int	  idx_end   = (ch+1)*AVC8K68XX_DMA_DESC_UNIT-1;

	dvprintk(DPRT_LEVEL0, dev, "%s(%d) is4CIF=%d\n", __func__, ch, is4CIF);

	/*
	 * 0 -- scatter gather, 2 -- frame block mode, 3 -- field block mode
	 */
	uTemp = (!IS_AVC8K68XX_DMA_MODE_BLOCK) ? 0 : (!is4CIF ? 3 : 2);

	uConfig  = avc8k_read( AVC8K68XX_R_PHASE_REF_CONFIG );
	uConfig &= ~(3<<(16+(ch<<1)));
	uConfig |= (uTemp<<(16+(ch<<1)));
	avc8k_write( AVC8K68XX_R_PHASE_REF_CONFIG, uConfig );

	uConfig = avc8k_read(AVC8K68XX_R_CH0to7_CONFIG_REG(ch));
	uConfig = (uConfig & ~0x1FF)		| (idx_begin&0x1FF);			//m_StartIdx
	uConfig = (uConfig & ~(0x3FF<<10))	| ((idx_end&0x3FF)<<10);		//m_EndIdx
	uConfig = (uConfig & ~(1<<23))		| ((false&1)<<23);				//m_bHorizontalDecimate
	uConfig = (uConfig & ~(1<<24))		| ((false&1)<<24);				//m_bVerticalDecimate
	uConfig = (uConfig & ~(3<<25))		| ((0&3)<<25);					//m_nDropChannelNum
	uConfig = (uConfig & ~(1<<27))		| ((1&1)<<27);					//m_bDropMasterOrSlave	//((((ch&1)?FALSE:TRUE)&1)<<27)
	uConfig = (uConfig & ~(1<<28))		| ((false&1)<<28);				//m_bDropField			//((bDropField&1)<<28)
	uConfig = (uConfig & ~(1<<29))		| ((false&1)<<29);				//m_bDropOddOrEven
	//uConfig = (uConfig & ~(3<<30))	| ((0&3)<<30);					//m_nCurVideoChannelNum -- mux

	avc8k_write(AVC8K68XX_R_CH0to7_CONFIG_REG(ch), uConfig);
	avc8k_write(AVC8K68XX_R_DROP_FIELD_REG(ch), 0);					        //0x95555555 for even only, 0xd5555555 for odd only

    if( !IS_AVC8K68XX_DMA_MODE_BLOCK ) {
#define DMA_STATUS_HOST_NOT_AVAIABLE    0
        dma_desc = AVC8K68XX_DMA_DESC(chip, P);  // 6869 uses dev instead of ch
        for( uTemp=0; uTemp<AVC8K68XX_DMA_DESC_UNIT; uTemp++ ) {
            dma_desc[uTemp].ctrl = (DMA_STATUS_HOST_NOT_AVAIABLE&3)<<30;
        }
        dma_desc = AVC8K68XX_DMA_DESC(chip, B); // 6869 uses dev instead of ch
        for(uTemp=0; uTemp<AVC8K68XX_DMA_DESC_UNIT; uTemp++) {
            dma_desc[uTemp].ctrl = (DMA_STATUS_HOST_NOT_AVAIABLE&3)<<30;
        }
    }
    else {
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_P_0, ch), chip->dma_bouncebuffer.dma/*0*/ );
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_B_0, ch), chip->dma_bouncebuffer.dma/*0*/ );
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_WHP_0, ch),
                    (linewidth&0x7FF) |						//line width
                    ((linepitch&0x7FF)<<11) |				//pitch
                    ((nHeight&0x7FF)<<22)					//height
                    );
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_P_F2_0, ch), chip->dma_bouncebuffer.dma/*0*/ );
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_B_F2_0, ch), chip->dma_bouncebuffer.dma/*0*/ );
        avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_WHP_F2_0, ch),
                    (linewidth&0x7FF) |						//line width
                    ((linepitch&0x7FF)<<11) |				//pitch
                    ((nHeight&0x7FF)<<22)					//height
                    );
    }

	//field 2
	if(!is4CIF) {
		avc8k_dev_set_size(dev, width, nHeight);
        avc8k_dev_set_size2(dev, width, nHeight);
	}
	else {
		avc8k_dev_set_size2(dev, width, nHeight);
		avc8k_dev_set_size(dev, width, nHeight);
	}
}

// We should probably use V4L's field setting to determine which buffer to read.
// Assume that if the P/B flag is set to point to a particular pair, both fields are valid.
#define USE_HW_FIELD
#ifdef USE_HW_FIELD // Use the hardware flag to select the buffer
#define SECOND_FIELD (status>>24)&1
#else // Use V4L's top/bottom field flag to select the buffer
#define SECOND_FIELD V4L2_FIELD_HAS_BOTTOM(dev->field)
#endif

int avc8k_dev_set_dmabuf(struct avc8k_vdev *dev, int nbuf, struct avc8k_buf *buf)
{
    struct avc8k_chip *chip = dev->chip;
    int ch = dev->channel_id;
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(3,11,0)) || (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)))
    dma_addr_t dma_addr = buf ? avc8k_videobuf_to_dma_contig(&buf->vb) : chip->dma_bouncebuffer.dma;
#else
    dma_addr_t dma_addr = buf ? videobuf_to_dma_contig(&buf->vb) : 0;
#endif

    dvprintk(DPRT_LEVEL0, dev, "%s(n_buf=%d buf=%d 0x%08x)\n", __func__, nbuf, buf->vb.i, (u32)dma_addr);

    if( V4L2_FIELD_HAS_BOTH(dev->field) ) {
		if((avc8k_read( AVC8K68XX_R_DMA_PB_STATUS )>>ch)&1) //Select P/B buffer
            avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_B_0, ch), (u32)dma_addr );
		else
            avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_P_0, ch), (u32)dma_addr );
    }
    else {
		u32 status=avc8k_read( AVC8K68XX_R_DMA_PB_STATUS )>>ch;
		if(status&1) //Select P/B buffer
		{
			if(SECOND_FIELD) //select field
				avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_B_F2_0, ch), (u32)dma_addr );
			else
				avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_B_0, ch), (u32)dma_addr );
		}
		else
		{
			if(SECOND_FIELD) //select field
				avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_P_F2_0, ch), (u32)dma_addr );
			else
				avc8k_write( AVC8K68XX_R_BDMA(AVC8K68XX_R_BDMA_ADDR_P_0, ch), (u32)dma_addr );
		}
    }

	return 0;
}

#define AVC8K_DISABLE_DESCRIPTOR(desc)     \
{                                           \
	(desc)->ctrl &= 0x3FFFFFFF;             \
	(desc)->ctrl |= 0x20000000;             \
}

#define AVC8K_FILL_DESCRIPTOR(desc, phy_addr, byte_len, field_start, next_start, new_frame, check_point, status)    \
{                                           \
	(desc)->addr = phy_addr;                \
	(desc)->ctrl = ((status & 0x3) << 30) | \
		((new_frame&1)<<29)				  | \
		((next_start & 0xFF) << 21)       | \
		((field_start & 0x7F) << 14)	  | \
		((check_point&1)<<13)			  | \
		(byte_len & 0x1FFF);                \
}

#define DMA_STATUS_HOST_READY			1

int avc8k_dev_set_dmadesc(struct avc8k_vdev *dev, int nbuf, struct avc8k_buf *buf)
{
    struct avc8k_chip *chip = dev->chip;
 //   int ch = dev->channel_id;
//	int chk_point = false;
	int fld_start = false;
	int i, j, bytes_to;
	int offset=0, n_desc=1;
	struct videobuf_dmabuf *dma = videobuf_to_dma(&buf->vb);
	struct avc8k_dmadesc  *desc;
	u32 phy_addr = 0;

	int fld_num = V4L2_FIELD_HAS_BOTH(buf->vb.field) + 1;
	int fld_cnt = 0;

    n_desc += (dev->pb_next==0);

 	dvprintk(DPRT_LEVEL0, dev, "%s(%d %d), %d %p %p %p, %x\n", __func__, nbuf, buf->vb.i,
             n_desc, AVC8K68XX_DMA_DESC(chip, P), AVC8K68XX_DMA_DESC(chip, B), buf->sg, buf->sg_offset);

    for( j=0; j<n_desc; j++ ) {
        if((buf->sg==NULL) || (buf->sg_bytes_pb==buf->vb.size)) {
            buf->sg = dma->sglist;
            buf->sg_offset = 0;
            buf->sg_bytes_to = buf->vb.size;
      //      chk_point = true;
            fld_start = true;
			fld_cnt = 0;
            offset= 0;
        }
        else {
            offset = buf->sg_offset;
        }
        bytes_to= buf->sg_bytes_pb/fld_num;
        desc  = (nbuf==0) ? AVC8K68XX_DMA_DESC(chip, P) : AVC8K68XX_DMA_DESC(chip, B);//6869 uses dev instead of ch
        nbuf  = (nbuf + 1) % dev->buf_needs;
        for(i=0; i<AVC8K68XX_DMA_DESC_UNIT; i++) {
            if(bytes_to == 0) {
                AVC8K_DISABLE_DESCRIPTOR(&desc[i]);
            }
            else {
                while (offset && offset >= sg_dma_len(buf->sg)) {
                    offset -= sg_dma_len(buf->sg);
                    buf->sg++;
                }
                phy_addr = cpu_to_le32(sg_dma_address(buf->sg)+offset);
                if(offset && (offset<AVC8K68XX_DMA_DESC_LEN)) {
                    offset = min(AVC8K68XX_DMA_DESC_LEN-offset, bytes_to);
                    AVC8K_FILL_DESCRIPTOR(&desc[i], phy_addr, offset, 0, 0, fld_start, 1, DMA_STATUS_HOST_READY);
                    bytes_to-= offset;
                    offset   = AVC8K68XX_DMA_DESC_LEN;
                    //dvprintk(1, dev, "%s+(%x %x)\n", (nbuf==0)?"B":"P", desc[i].ctrl, phy_addr);
					fld_start = false;
                }
                else if(bytes_to > AVC8K68XX_DMA_DESC_LEN) {
                    AVC8K_FILL_DESCRIPTOR(&desc[i], phy_addr, AVC8K68XX_DMA_DESC_LEN, 0,0,fld_start, 1, DMA_STATUS_HOST_READY);
                    bytes_to-= AVC8K68XX_DMA_DESC_LEN;
                    offset  += AVC8K68XX_DMA_DESC_LEN;
                    //dvprintk(1, dev, "%s (%x %x)\n", (nbuf==0)?"B":"P", desc[i].ctrl, phy_addr);
					fld_start = false;
                }
                else {
                    AVC8K_FILL_DESCRIPTOR(&desc[i], phy_addr, bytes_to, 0, 0,fld_start, 1, DMA_STATUS_HOST_READY);
                    offset  += bytes_to;
                    bytes_to = 0;
                    //dvprintk(1, dev, "%s-(%x %x)\n", (nbuf==0)?"B":"P", desc[i].ctrl, phy_addr);
		// -r 6869
		//		}
        //        fld_start = false;
					fld_cnt++;
                    if(fld_cnt < fld_num)
                    {
                        fld_start = true;
                        bytes_to  = buf->sg_bytes_pb/fld_num;
                    }
                }

            }
        }
        buf->sg_offset = offset;
 //       chk_point = false;  -r 6869
    }

    return 0;
}

//-------------------------------------------------------------------
// Allows GPIO output
//-------------------------------------------------------------------
bool avc8k_dev_enable_gpiooutput(struct avc8k_chip *chip, bool enable, int gpio, int bits)
{
	u32 mask= ((1 << bits) - 1) << (gpio+16);
	u32 old = avc8k_read(AVC8K68XX_R_GPIO_REG);
	avc8k_write(AVC8K68XX_R_GPIO_REG, enable ? (old&~mask) : (old|mask));

	dprintk(DPRT_LEVEL0, chip, "%s()\n", __func__);

	return true;
}

//-------------------------------------------------------------------
// Output GPIO level
//-------------------------------------------------------------------
bool avc8k_dev_set_gpiobits(struct avc8k_chip *chip, int gpio, int bits, u32* value)
{
	u32 mask= ((1 << bits) - 1) << gpio;
	u32 old = avc8k_read(AVC8K68XX_R_GPIO_REG) & ~mask;
	avc8k_write(AVC8K68XX_R_GPIO_REG, old|((*value<<gpio) & 0xFFFF0000));

	dprintk(DPRT_LEVEL0, chip, "%s()\n", __func__);

	return true;
}

//-------------------------------------------------------------------
// Access GPIO
//-------------------------------------------------------------------
bool avc8k_dev_get_gpiobits(struct avc8k_chip *chip, int gpio, int bits, u32* pValue)
{
	*pValue = (((avc8k_read(AVC8K68XX_R_GPIO_REG) & 0xFFFF0000) & (((1 << bits) - 1) << gpio)) >> gpio);

	dprintk(DPRT_LEVEL0, chip, "%s()\n", __func__);

	return true;
}

int avc8k_dev_set_audio(struct avc8k_chip *chip, int samplerate, int bits, int channels, int blksize)
{
  u32 tmp1, tmp2, tmp3;

  tmp1 = avc8k_read(AVC8K68XX_R_AUDIO_CTRL1);

  tmp1 &= 0x0000001F;
  tmp1 |= (125000000/samplerate) << 5;
  tmp1 |= blksize << 19;

  avc8k_write(AVC8K68XX_R_AUDIO_CTRL1,  tmp1);

  tmp2 = ((125000000 / samplerate) << 16) +
         ((125000000 % samplerate) << 16)/samplerate;

  avc8k_write(AVC8K68XX_R_AUDIO_CTRL2,  tmp2);

  tmp3 = avc8k_read(AVC8K68XX_R_AUDIO_CTRL3);
  tmp3 &= ~BIT(8);
  tmp3 |= (bits==8) ? BIT(8) : 0;
  avc8k_write(AVC8K68XX_R_AUDIO_CTRL3, tmp3);

  return 0;
}

int avc8k_dev_set_adma_buffer(struct avc8k_adev *dev, u32 buf_addr, int pb)
{
    struct avc8k_chip *chip = dev->chip;

    if(pb == 0) {
        avc8k_write(AVC8K68XX_R_CH8to15_CONFIG_REG_P(dev->channel_id), buf_addr);
    }
    else {
        avc8k_write(AVC8K68XX_R_CH8to15_CONFIG_REG_B(dev->channel_id), buf_addr);
    }

    return 0;
}

//----------------------------------------------
// Start/Stop Audio DMA program
//----------------------------------------------
void avc8k_dev_run_audio(struct avc8k_adev *dev, bool b_run)
{
    struct avc8k_chip *chip = dev->chip;
    int  ch   = dev->channel_id+AVC8K68XX_AUDIO_BEGIN;
	u32  tmp  = avc8k_read(AVC8K68XX_R_DMA_CMD);
	u32  tmp1 = avc8k_read(AVC8K68XX_R_DMA_CHANNEL_ENABLE);

	dvprintk(DPRT_LEVEL0, dev, "%s(%d %d)\n", __func__, ch, b_run);

	if(b_run) {
        tmp |= (1<<31);
        tmp |= (1<<ch);
        tmp1|= (1<<ch);
	}
	else {
		if(!(tmp&(1<<ch)) && !(tmp1&(1<<ch))) {
			return;
		}
		tmp &= ~(1<<ch);
		tmp1&= ~(1<<ch);
		if( tmp1 == 0 ) {
		    tmp = 0;
		}
	}

	avc8k_write(AVC8K68XX_R_DMA_CHANNEL_ENABLE, tmp1);
	avc8k_write(AVC8K68XX_R_DMA_CMD, tmp);
}


//----------------------------------------------
// Set dma p/b buffer desc
//----------------------------------------------
void avc8k_dev_set_pbdesc(struct avc8k_vdev *dev)
{
    struct avc8k_chip *chip = dev->chip;

	u32 page_table_p[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE0, AVC8K6869_R_DMA_PAGE_TABLE2, AVC8K6869_R_DMA_PAGE_TABLE4, AVC8K6869_R_DMA_PAGE_TABLE6,
		AVC8K6869_R_DMA_PAGE_TABLE8, AVC8K6869_R_DMA_PAGE_TABLEA, AVC8K6869_R_DMA_PAGE_TABLEC, AVC8K6869_R_DMA_PAGE_TABLEE
	};
	u32 page_table_b[AVC8K68XX_MAX_NUM_SG_DMA] = {
		AVC8K6869_R_DMA_PAGE_TABLE1, AVC8K6869_R_DMA_PAGE_TABLE3, AVC8K6869_R_DMA_PAGE_TABLE5, AVC8K6869_R_DMA_PAGE_TABLE7,
		AVC8K6869_R_DMA_PAGE_TABLE9, AVC8K6869_R_DMA_PAGE_TABLEB, AVC8K6869_R_DMA_PAGE_TABLED, AVC8K6869_R_DMA_PAGE_TABLEF
	};

	avc8k_write(page_table_p[dev->channel_id], (u32)chip->dma_desc[AVC8K68XX_DMA_DESC_P].dma);
	avc8k_write(page_table_b[dev->channel_id], (u32)chip->dma_desc[AVC8K68XX_DMA_DESC_B].dma);
}

int SendDataBit(struct avc8k_chip *chip, int b)
{
	unsigned int control_word;
	control_word = 0;

	CLR_SERIAL_CLK;		//si assicura che SCK low
	if(b)
	{
		GET_DOUT;
	}
	else
	{
		CLOSE_OUT;
	}

	DELAY_UNIT;

	SET_SERIAL_CLK;		//device latch data bit now!

	DELAY_UNIT;

	CLR_SERIAL_CLK;

	return 0;
}

// ritorna un numero negativo se c'e` un errore, altrimenti 0 o 1
int RecDataBit(struct avc8k_chip *chip)
{
	int b;
	unsigned int  control_word;

	control_word = 0;

	CLR_SERIAL_CLK;		//si assicura che SCK low

	DELAY_UNIT;

	SET_SERIAL_CLK;

	DELAY_UNIT;

	b = ((DeviceReadRegister(chip,CLK_REG))>>(GPIO_DIN_BIT))&1;
	CLR_SERIAL_CLK;

	return b;
}


int SendDataWord(struct avc8k_chip *chip, int wo, int wlen)
{
	int k;
	unsigned int  control_word;
//	CCaptureDevice *m_Device = dev;

	control_word = 0;

	CLR_SERIAL_CLK;
	for (k = wlen-1; k >= 0; k--)
		SendDataBit(chip,wo & (1<<k));

	CLOSE_OUT;

	return 0;
}

//Standard Receive data word
int RecDataWord(struct avc8k_chip *chip, int wlen)
{
	int k, val = 0;
	unsigned int  control_word;
//	CCaptureDevice *m_Device = dev;

	control_word = 0;

	CLR_SERIAL_CLK;
	for (k = wlen-1; k >= 0; k--)
		if ( RecDataBit(chip) )
			val |= (1<<k);

	return val;
}


int WaitReadyAfterWrite(struct avc8k_chip *chip)
{
	unsigned int  control_word;
	long k;
	CLR_SERIAL_CLK;
	SET_SENSOR2;

	DELAY_UNIT;

	SET_SENSOR1;

	DELAY_UNIT;

	CLR_SERIAL_CLK;


	for (k = 50000; k > 0  &&  (!(((DeviceReadRegister(chip,CLK_REG))>>(GPIO_DIN_BIT))&1)); k--)
		DELAY_UNIT;		//07/08/99 ** try to fix temporization (so to need only one global calibration)

	return k ? 0 : -1;
}

void GPIO_CTAL(struct avc8k_chip *chip,unsigned int GPIOaddr)
{
	unsigned int  GPIO_C_ADDR, control_word;

	GPIO_C_ADDR=GPIOaddr;
	control_word = 0;

	control_word = DeviceReadRegister(chip,GPIO_C_ADDR);

	control_word &= ~(1<<(TEC_CLK_BIT+16));
	control_word &= ~(1<<(TEC_DOUT_BIT+16));
	control_word &= ~(1<<(GPIO_CS_BIT+16));
	control_word |= (1<<(GPIO_DIN_BIT+16));
	control_word |= (1<<TEC_CLK_BIT);
	control_word |= (1<<TEC_DOUT_BIT);
	DeviceWriteRegister(chip,GPIO_C_ADDR, control_word);

//	printf("SPI control write GPIO 0x%X control_word= 0X%x \n", GPIOaddr, control_word );
//	printf("SPI control by GPIO 0X%X control_word= 0X%x \n", GPIOaddr, control_word );
	///udelay(I2C_WAIT_DELAY);  // 5us ?
//	usleep(5);

	SET_SENSOR2
	SET_SENSOR1
	SendCmdOpcode(chip,WriteEnableCode);
	SendDataWord(chip,0xFFFF, address_len);
	SET_SENSOR2
}


unsigned char DeviceReadFPGA1(struct avc8k_chip *chip,unsigned long  byte_addr)
{
	unsigned int  control_word;
	unsigned long	Data;

	SET_SENSOR1

	SendCmdOpcode(chip,ReadCode);
	SendAddress(chip,byte_addr, address_len);

	Data = RecDataWord(chip,data_len);

	SET_SENSOR2
	DELAY_UNIT;

	return (unsigned char)Data;
}



unsigned long DeviceReadRegister(struct avc8k_chip *chip,unsigned long addr)
{
	return avc8k_read(addr);
}
void DeviceWriteRegister(struct avc8k_chip *chip,unsigned long addr, unsigned long value)
{
	avc8k_write(addr,value);
}
