/*
 * AVC8000 Copyright (C) 2012 Advanced Micro Peripherals Ltd
 * Driver for AVC8000 cards
 *
 * (c) 2009-10 liran <jlee@techwellinc.com.cn> [Techwell China]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include "avc8000.h"
#include "avc8000-reg.h"


static bool avc8k_i2c_s_scl( struct i2c_sw *i2c, TLevel scl )
{
    struct avc8k_chip *chip = (struct avc8k_chip *)i2c->chip;
    struct avc8k_i2c *i2c_686x = &chip->i2c_686x;

	i2c_686x->control_word &= ~AVC8K6864_I2C_SCLK_CTRL_BIT;
	if(scl == LevelHi) {
		i2c_686x->control_word |= AVC8K6864_I2C_SCLK_BIT;
	}
	else {
		i2c_686x->control_word &= ~AVC8K6864_I2C_SCLK_BIT;
	}

	avc8k_write(AVC8K68XX_R_GPIO_REG, i2c_686x->control_word);

	return i2csw_s_scl(&i2c_686x->i2c, scl);
}

static bool avc8k_i2c_s_sda( struct i2c_sw *i2c, TLevel sda )
{
    struct avc8k_chip *chip = (struct avc8k_chip *)i2c->chip;
    struct avc8k_i2c *i2c_686x = &chip->i2c_686x;

	i2c_686x->control_word &= ~AVC8K6864_I2C_SDA_CTRL_BIT;
	if(sda == LevelHi) {
		i2c_686x->control_word |= AVC8K6864_I2C_SDA_DATA_BIT;
	}
	else {
		i2c_686x->control_word &= ~AVC8K6864_I2C_SDA_DATA_BIT;
	}

	avc8k_write(AVC8K68XX_R_GPIO_REG, i2c_686x->control_word);

	return i2csw_s_sda(&i2c_686x->i2c, sda);
}

static int avc8k_i2c_g_scl( struct i2c_sw * i2c )
{
    struct avc8k_chip *chip = (struct avc8k_chip *)i2c->chip;

	return (avc8k_read(AVC8K68XX_R_GPIO_REG) & AVC8K6864_I2C_SCLK_BIT);
}

static int avc8k_i2c_g_sda( struct i2c_sw * i2c )
{
    struct avc8k_chip *chip = (struct avc8k_chip *)i2c->chip;
    struct avc8k_i2c *i2c_686x = &chip->i2c_686x;

	i2c_686x->control_word |= AVC8K6864_I2C_SDA_CTRL_BIT;
	avc8k_write(AVC8K68XX_R_GPIO_REG, i2c_686x->control_word);

	return ((avc8k_read(AVC8K68XX_R_GPIO_REG) & AVC8K6864_I2C_SDA_DATA_BIT)>>1);
}

static int i2c_xfer(struct i2c_adapter *i2c_adap,
		    struct i2c_msg *msgs, int num)
{
	struct avc8k_chip *chip = i2c_adap->algo_data;
	struct i2c_sw       *i2c = &chip->i2c_686x.i2c;
	int i, retval = 0;

	dprintk(1, chip, "%s(num = %d)\n", __func__, num);

	for (i = 0 ; i < num; i++) {
		dprintk(1, chip, "%s(num = %d) addr = 0x%02x  len = 0x%x\n",
			__func__, num, msgs[i].addr, msgs[i].len);
		if (msgs[i].flags & I2C_M_RD) {
			/* read */
			retval = i2csw_read_bytes(i2c, (u8*)&msgs[i].addr, 1, msgs[i].buf, msgs[i].len);
		} else {
			/* write */
			retval = i2csw_write_bytes(i2c, msgs[i].buf, msgs[i].len);
		}
		if (!retval) {
		    retval = -i2c->err_code;
			goto err;
		}
	}
	return num;

 err:
	return retval;
}

static u32 avc8k_functionality(struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_EMUL | I2C_FUNC_I2C;
}

static struct i2c_algorithm avc8k_i2c_algo_template = {
	.master_xfer	= i2c_xfer,
	.functionality	= avc8k_functionality,
};

static struct i2c_adapter avc8k_i2c_adap_template = {
	.name              = "avc8000",
	.owner             = THIS_MODULE,
#if(LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39))
	.id                = I2C_SW_AVC8K64XX,
#endif
	.algo              = &avc8k_i2c_algo_template,
};

static struct i2c_client avc8k_client_template = {
	.name	= "avc8000 internal",
};

static struct i2c_sw avc8k_i2csw_template = {
    .s_scl = avc8k_i2c_s_scl,
    .s_sda = avc8k_i2c_s_sda,
    .g_scl = avc8k_i2c_g_scl,
    .g_sda = avc8k_i2c_g_sda,
};

static int avc8k_i2c_init( struct avc8k_chip *chip )
{
    struct avc8k_i2c *i2c_686x = &chip->i2c_686x;

    i2c_686x->i2c = avc8k_i2csw_template;
	i2c_686x->i2c.chip = chip;

	i2c_686x->control_word = avc8k_read(AVC8K68XX_R_GPIO_REG);
	i2c_686x->control_word &= ~AVC8K6864_I2C_SCLK_CTRL_BIT;
	i2c_686x->control_word &= ~AVC8K6864_I2C_SDA_CTRL_BIT;
	i2c_686x->control_word |= AVC8K6864_I2C_SCLK_BIT;
	i2c_686x->control_word |= AVC8K6864_I2C_SDA_DATA_BIT;
	avc8k_write(AVC8K68XX_R_GPIO_REG, i2c_686x->control_word);

    i2csw_init( &i2c_686x->i2c, 400000);//2000000);

    return 0;
}

/* init + register i2c algo-bit adapter */
int avc8k_i2c_register(struct avc8k_chip *chip)
{
	dprintk(1, chip, "%s()\n", __func__);

	chip->i2c_adap = avc8k_i2c_adap_template;
	chip->i2c_adap.dev.parent = &chip->pci->dev;
	strcpy(chip->i2c_adap.name, chip->name);
	chip->i2c_adap.algo_data = chip;
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	i2c_set_adapdata(&chip->i2c_adap, &chip->v4l2_dev);
#else
	i2c_set_adapdata(&chip->i2c_adap, chip);
#endif
	i2c_add_adapter(&chip->i2c_adap);

	chip->i2c_client = avc8k_client_template;
	chip->i2c_client.adapter = &chip->i2c_adap;
	chip->i2c_client.addr = AVC8K2864_R_I2C_ADDR0;

    avc8k_i2c_init( chip );
	return 0;
}

int avc8k_i2c_unregister(struct avc8k_chip* chip)
{
	i2c_del_adapter(&chip->i2c_adap);
	return 0;
}

unsigned char DeviceWriteFPGA1(struct avc8k_chip *chip, unsigned long  byte_addr)
{
	unsigned int  GPIO_C_ADDR;
	unsigned char	Data[3];
	struct i2c_sw *i2c;

	GPIO_C_ADDR = CLK_REG;

	i2c = &chip->i2c_686x.i2c;

	Data[0] = 0xA0;

	Data[1] = (unsigned char)(byte_addr&0xFF);	
	Data[0] = 0xA0;

	if(!i2csw_write_bytes(i2c, Data, 2)) {
		return 0xFF;
	}

	
	Data[0] |= 1;	

	if(!i2csw_read_bytes(i2c, Data, 1, &Data[2], 1)) {
		return 0xFF;
	}

	return Data[2];
}
