/*
 * AVC8000 Copyright (C) 2012 Advanced Micro Peripherals Ltd
 * Driver for AVC8000 cards
 *
 * (c) 2009-10 liran <jlee@techwellinc.com.cn> [Techwell China]
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef AVC8000_H_INCLUDED
#define AVC8000_H_INCLUDED




#include <linux/pci.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/videodev2.h>
#include <linux/kdev_t.h>
#include <linux/input.h>
#include <linux/notifier.h>
#include <linux/delay.h>
#include <linux/mutex.h>


#include <asm/io.h>
#include <linux/version.h>

#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,38))
#include <linux/module.h>
#endif

#include <media/v4l2-common.h>
#include <media/v4l2-ioctl.h>
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
#include <media/v4l2-device.h>
#endif
#include <media/videobuf-dma-sg.h>
#include <sound/core.h>
#include <sound/pcm.h>
//#include "avc8000-videobuf-dma-contig.h"

#define AVC8K_VERSION_CODE KERNEL_VERSION(1, 0, 14)

#define  PCI_DEVICE_ID_AVC8K6869		0x6869
#define  PCI_DEVICE_ID_AVC8K6864		0x6864

// dummy RHEL version define when used with standard kernel
#ifndef RHEL_RELEASE_CODE
#define RHEL_RELEASE_CODE 0
#define RHEL_RELEASE_VERSION(a,b) (a + b)
#endif
//



#include "avc8000-i2c-sw.h"

#define I2C_SW_AVC8K64XX	0x0A0000 /* AVC8000 video decoder bus */

extern unsigned int avc8k_bcsi;	/* for BCSI */
extern unsigned int avc8k_restart_timer;	/* dma restart timer */
extern unsigned int avc8k_dmamode;
extern unsigned int avc8k_debug;
#define dprintk(level, chip, fmt, arg...)\
	do { if (avc8k_debug >= level)\
		printk(KERN_INFO "%s: " fmt, (chip)->name, ## arg);\
	} while (0)
#define dvprintk(level, dev, fmt, arg...)\
	do { if (avc8k_debug >= level)\
		printk(KERN_INFO "%s[v-%d]: " fmt, ((dev)->chip)->name, (dev)->channel_id, ## arg);\
	} while (0)
#define daprintk(level, dev, fmt, arg...)\
	do { if (avc8k_debug >= level)\
		printk(KERN_INFO "%s[a-%d]: " fmt, ((dev)->chip)->name, (dev)->channel_id, ## arg);\
	} while (0)

#define dvcount(level, dev, cnt, cnt_out)       \
{                                               \
    if(avc8k_debug >= level) {                 \
        static struct timeval _tv_begin_##cnt = {0,0};  \
        static long _cnt_##cnt = 0;             \
        struct timeval _tv_now_##cnt;           \
        long long _tv_diff_##cnt = 0;           \
        long _ms_diff_##cnt = 0;                \
        do_gettimeofday( &_tv_now_##cnt );      \
        _tv_diff_##cnt = 1000000 * ( _tv_now_##cnt.tv_sec - _tv_begin_##cnt.tv_sec ) +      \
                  _tv_now_##cnt.tv_usec - _tv_begin_##cnt.tv_usec;                          \
        _ms_diff_##cnt = (long)_tv_diff_##cnt;  \
        _ms_diff_##cnt /= 1000;                 \
        _cnt_##cnt++;                           \
        if(_ms_diff_##cnt >= 1000) {            \
            dvprintk(1, dev, "%s-%d %d\n", cnt_out, (int)_cnt_##cnt, (int)_ms_diff_##cnt);  \
            _tv_begin_##cnt = _tv_now_##cnt;    \
            _cnt_##cnt = 0;                     \
        }                                       \
    }                                           \
}

#define DPRT_LEVEL0        1

#define RESOURCE_VIDEO     2
#define BUFFER_TIMEOUT     (HZ)  /* 0.5 seconds */

/* ioctl for avc8000 */
#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
#define AVC8K68XX_S_REGISTER           _IOWR('V', (BASE_VIDIOC_PRIVATE + 1), struct v4l2_dbg_register)
#define AVC8K68XX_G_REGISTER           _IOWR('V', (BASE_VIDIOC_PRIVATE + 2), struct v4l2_dbg_register)
#else
#define AVC8K68XX_S_REGISTER           _IOWR('V', (BASE_VIDIOC_PRIVATE + 1), struct v4l2_register)
#define AVC8K68XX_G_REGISTER           _IOWR('V', (BASE_VIDIOC_PRIVATE + 2), struct v4l2_register)
#endif
#define AVC8K68XX_S_FRAMERATE          _IOWR('V', (BASE_VIDIOC_PRIVATE + 3), long)
#define AVC8K68XX_G_FRAMERATE          _IOWR('V', (BASE_VIDIOC_PRIVATE + 4), long)
#define AVC8K68XX_G_SIGNAL             _IOWR('V', (BASE_VIDIOC_PRIVATE + 5), long)

#define AVC8K68XX_DAEMON_PERIOD        2000    /*2s*/
#define AVC8K68XX_DAEMON_THRESHOLD     16      /*2s*/
#define AVC8K68XX_DAEMON_TOLERANCE     2       /*2 times*/
#define AVC8K68XX_DAEMON_MIN           2       /*min frame rate*/

#define AVC8K68XX_DMA_MODE_SG          0       /* 0--scatter gather, 1--block dma*/
#define AVC8K68XX_DMA_MODE_BLOCK       1
#define IS_AVC8K68XX_DMA_MODE_BLOCK    (avc8k_dmamode & AVC8K68XX_DMA_MODE_BLOCK)
#define AVC8K68XX_DMA_DESC_LEN         4096
#define AVC8K68XX_DMA_DESC_P           0
#define AVC8K68XX_DMA_DESC_B           1
#define AVC8K68XX_DMA_DESC_NUM         2
#define AVC8K68XX_DMA_DESC_UNIT		512//64 -r form 6869
#define AVC8K68XX_DMA_DESC_UNIT_LEN	(AVC8K68XX_DMA_DESC_UNIT*8)
//#define AVC8K68XX_DMA_DESC(ch, pb)     ((struct avc8k_dmadesc*)((unsigned long)chip->dma_desc[AVC8K68XX_DMA_DESC_##pb].cpu +((ch)*AVC8K68XX_DMA_DESC_UNIT_LEN))) */
#define AVC8K68XX_DMA_DESC(chip, pb)     ((struct avc8k_dmadesc*)((unsigned long)chip->dma_desc[AVC8K68XX_DMA_DESC_##pb].cpu))
#define AVC8K68XX_SG_PBLEN             (AVC8K68XX_DMA_DESC_UNIT*AVC8K68XX_DMA_DESC_LEN)

#define AVC8K68XX_DECODER_COUNT(chip)  ((chip == 0x6869) ? 8 : 4)      //8
#define AVC8K6864_MUX_COUNT            1
#define AVC8K2865_MUX_COUNT            1
#define AVC8K2864_MUX_COUNT            1
#define AVC8K68XX_VIDEO_COUNT          8 //16
#define AVC8K68XX_AUDIO_COUNT          8
#define AVC8K68XX_AUDIO_BEGIN          8
#define AVC8K68XX_BUFFER_COUNT         9
#define AVC8K68XX_DEFAULT_STANDARD     V4L2_STD_PAL	//V4L2_STD_NTSC
#define AVC8K68XX_DEFAULT_WIDTH        720// 704
#define AVC8K68XX_DEFAULT_HEIGHT       576// 480
#define AVC8K68XX_REMOVE_BLACKSTRIPE   (1 << 31) // 0
#define AVC8K68XX_VIDEO_SIZE(w,h)		((w) | ((h)<<16) | (1<<31))
#define AVC8K68XX_VIDEO_SIZE_F2(w,h)	((w) | ((h)<<16))	//Field2 size,  Available only for Rev.B or later
#define AVC8K68XX_MAX_FRAMERATE(std)   (IS_PAL(std) ? 25 : 30)
#define AVC8K68XX_MIN_WIDTH            160
#define AVC8K68XX_MIN_HEIGHT           120

/* Currently unsupported by the driver: PAL/H, NTSC/Kr*/
#define AVC8K68XX_NORMS                (V4L2_STD_NTSC |  V4L2_STD_PAL)

#define IS_PAL( std )               ( (std) & V4L2_STD_PAL )

struct avc8k_fmt {
	char  *name;
	u32   fourcc;          /* v4l2 format id */
	int   depth;
	int   flags;
};

struct avc8k_dmadesc {
	u32	  ctrl;   //DO NOT CHANGE SEQUENCE, and it is valid for little endian target PC only
	u32	  addr;
};

struct avc8k_fh {
	struct avc8k_vdev         *dev;
	enum v4l2_buf_type         type;
	u32                        resources;
	struct videobuf_queue      vidq;

	/* video capture */
	struct avc8k_fmt         *fmt;
	int                        width;
	int                        height;
	int                        bytesperline;
	int                        ctl_bright;
	int                        ctl_contrast;
	int                        ctl_hue;
	int                        ctl_saturation;
};

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
struct avc8k_pgtable {
	unsigned int                size;
	__le32                      *cpu;
	dma_addr_t					dma;
};
#endif

/* buffer for one video frame */
struct avc8k_buf {
	/* common v4l buffer stuff -- must be first */
	struct videobuf_buffer     vb;

	/* avc8k specific */
	struct avc8k_fmt*         fmt;
	struct list_head           active;

	int                        sg_bytes_to;
	int                        sg_bytes_pb;
	struct scatterlist         *sg;
	int                        sg_offset;
	int                        sg_ofo;      /*one p/b buffer for one frame*/

	int (*activate)(struct avc8k_vdev *dev, struct avc8k_buf *buf,
                    int dma_pb, bool queued);
};

struct avc8k_dmaqueue {
	struct list_head           active;
	struct list_head           queued;
	struct timer_list          timeout;
};

/* buffer for dma */
struct avc8k_dmabuf {
	unsigned int   size;
	__le32         *cpu;
	dma_addr_t     dma;
	struct vm_area_struct *vma;
};

struct avc8k_i2c {
    struct i2c_sw              i2c;
    u32                        control_word;
};

/*
 * audio main chip structure
 */
typedef struct snd_card_avc8k {
	struct snd_card *card;
	struct pci_dev *pci;
	struct avc8k_adev *dev;

	u16 mute_was_on;

	spinlock_t lock;
} snd_card_avc8k_t;

struct avc8k_chip {
	struct list_head           chiplist;

	/* pci i/o */
	char                       name[32];
	int                        nr;

#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
	struct v4l2_device         v4l2_dev;
#endif
	struct pci_dev             *pci;
	__u8                       __iomem *bmmio;

	/* i2c i/o */
	struct i2c_adapter         i2c_adap;
	struct i2c_client          i2c_client;
	struct avc8k_i2c          i2c_686x;

    struct avc8k_vdev         *vid_dev[AVC8K68XX_VIDEO_COUNT];
	int                        vid_count;
	int						   vid_ch;
	struct timeval             daemon_tv;
	int                        daemon_nr;
	int                        daemon_tor;
	struct avc8k_dmabuf       dma_desc[AVC8K68XX_DMA_DESC_NUM];

    struct avc8k_adev         *aud_dev[AVC8K68XX_AUDIO_COUNT];
	int                        aud_count;

    //assist timer, for video decoder status checking
	struct timer_list          tm_assist;

	// additional lock for compatibility with kernel 3.11
	struct mutex               lock;

	struct avc8k_dmabuf        dma_bouncebuffer;
};

/* global device status */
struct avc8k_vdev {
    struct avc8k_chip*        chip;

	struct list_head           devlist;
	struct mutex               lock;
	spinlock_t                 slock;

	/* various device info */
	unsigned int               resources;
	struct video_device        *video_dev;
	struct avc8k_dmaqueue     video_q;
	struct avc8k_dmabuf       dma_vb[AVC8K68XX_BUFFER_COUNT];
	struct timer_list          tm_restart;

	/* various v4l controls */
	v4l2_std_id	               tvnorm;
	unsigned int               input;

	/* other global state info */
	u32                        vstatus;

	/* for dma buffer queue */
	int                        buf_needs;
	int                        pb_next;
	int                        pb_curr;
	int                        dma_started;
	int                        channel_id;
	int                        field;
	int                        framerate;
};

#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
struct avc8k_dmasound {
	struct avc8k_pgtable	   pt;
	void                       *vaddr;
	struct scatterlist         *sglist;
	int                        sglen;
	int                        nr_pages;
};
#endif

/* dmasound dsp status */
struct avc8k_adev {
    struct avc8k_chip*        chip;
	int                        channel_id;
    snd_card_avc8k_t *        card;

	struct mutex               lock;
	spinlock_t                 slock;
	int                        minor_dsp;
	unsigned int               users_dsp;

	/* dsp */
	unsigned int               rate;
	unsigned int               channels;
	unsigned int               dma_running;
	unsigned int               blocks;
	unsigned int               blksize;
	unsigned int               bufsize;
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
	struct avc8k_dmasound	   dmasound;
#else
	struct videobuf_dmabuf     dma;
#endif
	unsigned int               dma_blk;
	unsigned int               read_offset;
	struct snd_pcm_substream   *substream;

    unsigned int               dma_blk_addr;
    unsigned int               pb_flag;
    unsigned int               period_idx;
    unsigned int               sample_bits;


};

#if(LINUX_VERSION_CODE > KERNEL_VERSION(2,6,30))
static inline struct avc8k_chip *to_avc8k(struct v4l2_device *v4l2_dev)
{
	return container_of(v4l2_dev, struct avc8k_chip, v4l2_dev);
}
#endif

/* ----------------------------------------------------------- */

static inline u32 avc8k_readl(const volatile void __iomem *addr, u32 reg)
{
	u32 ret = readl(addr);
	if(ret==0xffffffff)
	{
		//printk(KERN_INFO "*** READ 0xFFFFFFFF *** reg%08x\n", reg);
		printk(KERN_INFO "*** reg%08x = 0x%08x\n", reg, ret);
	}
	return ret;
}

static inline void avc8k_writel(u32 val, volatile void __iomem *addr, u32 reg)
{
	u32 ret;
	writel(val,addr);
	ret = readl(addr);
	//if(ret!=val)
	if(ret==0xffffffff)
	{
		//printk(KERN_INFO "*** READ 0xFFFFFFFF *** reg%08x\n", reg);
		printk(KERN_INFO "*** reg%08x = 0x%08x not 0x%08x\n", reg, ret, val);
	}
}

//#define avc8k_read(reg)             readl(chip->bmmio + (reg))
#define avc8k_read(reg)             avc8k_readl(chip->bmmio + (reg), (reg))
//#define avc8k_write(reg, value)     writel((value), chip->bmmio + (reg))
#define avc8k_write(reg, value)     avc8k_writel((value), chip->bmmio + (reg), (reg))

#define avc8k_andor(reg, mask, value) \
  writel((readl(chip->bmmio+(reg)) & ~(mask)) |\
  ((value) & (mask)), chip->bmmio+(reg))

#define avc8k_set(reg, bit)          avc8k_andor((reg), (bit), (bit))
#define avc8k_clear(reg, bit)        avc8k_andor((reg), (bit), 0)

#define call_all(dev, o, f, args...) \
	v4l2_device_call_all(&dev->v4l2_dev, 0, o, f, ##args)

extern struct list_head avc8k_chiplist;

/* ----------------------------------------------------------- */
/* avc8000-core.c                                               */
int avc8k_dmabuf_alloc(struct avc8k_chip *chip,
		       struct avc8k_dmabuf *buf, unsigned int size);
void avc8k_dmabuf_free(struct avc8k_chip *chip, struct avc8k_dmabuf *buf);
#if(LINUX_VERSION_CODE >= KERNEL_VERSION(4,0,0))
void avc8k_pgtable_free(struct pci_dev *pci, struct avc8k_pgtable *pt);
#endif


/* ----------------------------------------------------------- */
/* avc8000-video.c                                              */
/* Video */
extern int  avc8k_video_register(struct avc8k_vdev *dev);
extern void avc8k_video_unregister(struct avc8k_vdev *dev);
extern int  avc8k_video_irq(struct avc8k_vdev *dev, u32 dma_status, u32 pb_status);
extern int  avc8k_video_daemon(struct avc8k_chip *chip, u32 dma_error);
extern int  avc8k_video_status(struct avc8k_vdev *dev);
extern int 	avc8k_video_resetdma(struct avc8k_chip* chip, u32 dma_cmd);

/* ----------------------------------------------------------- */
/* avc8000-alsa.c                                              */
/* Audio */
extern int  avc8k_audio_create(struct avc8k_adev *dev);
extern int  avc8k_audio_free(struct avc8k_adev *dev);
extern void avc8k_audio_irq(struct avc8k_adev *dev, u32 dma_status, u32 pb_status);

/* avc8000-i2c.c                                                */
extern int  avc8k_i2c_register(struct avc8k_chip *chip);
extern int  avc8k_i2c_unregister(struct avc8k_chip *chip);


extern void avc8k_reset(struct avc8k_chip *chip);


/* ----------------------------------------------------------- */
/* tv norms                                                    */

static unsigned int inline norm_maxw(v4l2_std_id norm)
{
	//return (norm & (V4L2_STD_MN & ~V4L2_STD_PAL_Nc)) ? 720 : 768;
	return 704;
}

static unsigned int inline norm_maxh(v4l2_std_id norm)
{
	return (norm & V4L2_STD_625_50) ? 576 : 480;
}

#define CLK_REG        0xbc//0x2F

#define USAGE_ADDR		0x18
#define USAGE_SIZE		4

#define TEC_DOUT_BIT			8
#define TEC_CLK_BIT				9
#define GPIO_DIN_BIT				10
#define GPIO_CS_BIT				11

#ifndef DELAY_UNIT
#define DELAY_UNIT udelay(20)
#endif

unsigned long DeviceReadRegister(struct avc8k_chip *chip,unsigned long);
void DeviceWriteRegister(struct avc8k_chip *chip,unsigned long , unsigned long );

#define SET_SERIAL_CLK {control_word = DeviceReadRegister(chip,CLK_REG);\
				control_word |= (1<<TEC_CLK_BIT);     \
				DeviceWriteRegister(chip,CLK_REG,control_word);}

#define CLR_SERIAL_CLK {control_word = DeviceReadRegister(chip,CLK_REG);\
				control_word &= ~(1<<TEC_CLK_BIT);    \
				DeviceWriteRegister(chip,CLK_REG,control_word);}

#define GET_DOUT {control_word = DeviceReadRegister(chip,CLK_REG);\
				 control_word |= (1<<TEC_DOUT_BIT);\
				DeviceWriteRegister(chip,CLK_REG,control_word);}

#define CLOSE_OUT {control_word = DeviceReadRegister(chip,CLK_REG);\
				 control_word &= ~(1<<TEC_DOUT_BIT);\
				 DeviceWriteRegister(chip,CLK_REG,control_word);}

#define SET_SENSOR1 {control_word = DeviceReadRegister(chip,CLK_REG);\
				 control_word |= (1<<GPIO_CS_BIT);     \
				 DeviceWriteRegister(chip,CLK_REG,control_word);}

#define SET_SENSOR2 {control_word = DeviceReadRegister(chip,CLK_REG);\
				 control_word &= ~(1<<GPIO_CS_BIT);     \
				 DeviceWriteRegister(chip,CLK_REG,control_word);}

#define SendCmdOpcode(chip,y) SendDataWord(chip,y,3)

#define SendAddress(chip,addr, alen) SendDataWord(chip,addr, alen)

unsigned char DeviceReadFPGA1( struct avc8k_chip *chip, unsigned long  byte_addr);
unsigned char DeviceWriteFPGA1(struct avc8k_chip *chip, unsigned long  byte_addr);
void GPIO_CTAL(struct avc8k_chip *chip, unsigned int GPIOaddr);

#define	ReadCode 06
#define	WriteCode 05
#define	WriteEnableCode 04
#define	EraseAllCode 04
#define	PrClearCode 07

#define address_len 7
#define data_len 8


#endif
